-- | Tiny templating engine
module Template where

import Data.Text
import Data.Text.Lazy.Encoding (encodeUtf8)
import Data.ByteString.Lazy (ByteString)
import Data.Binary
import Data.Functor (void)
import Data.Void
import GHC.Generics
import Text.Megaparsec
import Text.Megaparsec.Char
import Data.Aeson (Value(..))

import qualified Data.Text.Lazy      as LT
import qualified Data.Aeson          as Aeson
import qualified Data.Aeson.KeyMap   as KeyMap
import qualified Data.Aeson.Key      as Key
-- import qualified Data.HashMap.Strict as HashMap


data Chunk
  = Raw Text
  | Var [Text]
  | For Text [Text] Template
  | If [Text] Template
  deriving (Eq, Show, Generic, Binary)

type Template = [Chunk]
type Context = Aeson.Object

type Parser = Parsec Void Text


-- RENDERING

render :: Template -> Context -> LT.Text
render chunks ctx = foldMap (renderChunk ctx) chunks

lookupCtx :: [Text] -> Context -> Maybe Aeson.Value
lookupCtx keys = aux keys . Aeson.Object
  where
    aux :: [Text] -> Aeson.Value -> Maybe Aeson.Value
    aux []     v          = Just v
    aux (k:ks) (Object h) = KeyMap.lookup (Key.fromText k) h >>= aux ks
    aux _      _          = Nothing

renderChunk :: Context -> Chunk -> LT.Text
renderChunk ctx (Raw t) = LT.fromStrict t
renderChunk ctx (Var ks) =
  case Aeson.fromJSON <$> lookupCtx ks ctx of
    Just (Aeson.Success v) -> v
    _                      -> mempty
renderChunk ctx (For kn ks c) =
  case lookupCtx ks ctx of
    Just (Aeson.Array arr) ->
      foldMap (\v -> foldMap (renderChunk (KeyMap.insert (Key.fromText kn) v ctx)) c) arr
    _                      -> mempty
renderChunk ctx (If ks c) =
  case lookupCtx ks ctx of
    Just _ -> foldMap (renderChunk ctx) c
    _      -> mempty


-- PARSING

parseTemplate :: Text -> Maybe Template
parseTemplate = parseMaybe templateP

templateP :: Parser Template
templateP = many chunkP

chunkP :: Parser Chunk
chunkP = choice [ varP , forP , ifP , rawP ]
  where
    identP :: Parser Text
    identP = pack <$> ((:) <$> letterChar <*> many alphaNumChar)

    keysP :: Parser [Text]
    keysP = try $ "$" *> sepBy1 identP "."

    varP :: Parser Chunk
    varP = try $ Var <$> between ldelim rdelim
      (space *> keysP <* space)

    rawP :: Parser Chunk
    rawP = Raw . pack <$>
      someTill anySingle (eof <|> lookAhead ldelim)

    forP :: Parser Chunk
    forP = try do
      ldelim *> space *> string "for" *> space1
      key <- identP <* space1 <* "in" <* space1
      val <- keysP <* space <* rdelim
      chunks <- manyTill chunkP (try (between ldelim rdelim (space *> "end" *> space)))
      return $ For key val chunks

    ifP :: Parser Chunk
    ifP = try do
      ldelim *> space *> string "if" *> space1
      val <- keysP <* space <* rdelim
      chunks <- manyTill chunkP (try (between ldelim rdelim (space *> "end" *> space)))
      return $ If val chunks

    ldelim = void "{{"
    rdelim = "}}"
